import json
import os
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict

import redis
import yaml
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from tree_traverser import CompressedTree

app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

@app.get('/favicon.ico', include_in_schema=False)
async def favicon():
    return FileResponse("favicon.ico")


if "LOCAL_CACHE" in os.environ:
    print("Getting data from local file")

    base = Path(os.environ["LOCAL_CACHE"])
    with open(base / "compressed_tree.json", "r") as f:
        json_tree = f.read()


    with open(base / "language.yaml", "r") as f:
        mars_language = yaml.safe_load(f)["_field"]
    
else:
    print("Getting cache from redis")
    r = redis.Redis(host="redis", port=6379, db=0)
    json_tree = r.get('compressed_catalog')
    assert json_tree, "No compressed tree found in redis"
    mars_language = json.loads(r.get('mars_language'))

print("Loading tree from json")
c_tree = CompressedTree.from_json(json.loads(json_tree))

print("Partialy decompressing tree, shoud be able to skip this step in future.")
tree = c_tree.reconstruct_compressed_ecmwf_style()

print("Ready to serve requests!")

def request_to_dict(request: Request) -> Dict[str, Any]:
    # Convert query parameters to dictionary format
    request_dict = dict(request.query_params)
    for key, value in request_dict.items():
        # Convert comma-separated values into lists
        if "," in value:
            request_dict[key] = value.split(",")

    return request_dict

def match_against_cache(request, tree):
    if not tree: return {"_END_" : {}}
    matches = {}
    for k, subtree in tree.items():
        if len(k.split("=")) != 2:
            raise ValueError(f"Key {k} is not in the correct format")
        key, values = k.split("=")
        values = set(values.split(","))
        if key in request:
            if isinstance(request[key], list):
                matching_values = ",".join(request_value for request_value in request[key] if request_value in values)
                if matching_values:
                    matches[f"{key}={matching_values}"] = match_against_cache(request, subtree)
            elif request[key] in values:
                matches[f"{key}={request[key]}"] = match_against_cache(request, subtree)

    if not matches: return {k : {} for k in tree.keys()}
    return matches


def max_tree_depth(tree):
    "Figure out the maximum depth of a tree"
    if not tree:
        return 0
    return 1 + max(max_tree_depth(v) for v in tree.values())

def prune_short_branches(tree, depth = None):
    if depth is None: 
        depth = max_tree_depth(tree)
    return {k : prune_short_branches(v, depth-1) for k, v in tree.items() if max_tree_depth(v) == depth-1}
        
def get_paths_to_leaves(tree):
    for k,v in tree.items():
        if not v:
            yield [k,]
        else: 
            for leaf in get_paths_to_leaves(v):
                yield [k,] + leaf

def get_leaves(tree):
    for k,v in tree.items():
        if not v:
            yield k
        else: 
            for leaf in get_leaves(v):
                yield leaf

@app.get("/api/tree")
async def get_tree(request: Request):
    request_dict = request_to_dict(request)
    print(c_tree.multi_match(request_dict))
    return c_tree.multi_match(request_dict)

@app.get("/api/match")
async def get_match(request: Request):
    # Convert query parameters to dictionary format
    request_dict = request_to_dict(request)

    # Run the schema matching logic
    match_tree = match_against_cache(request_dict, tree)


    # Prune the tree to only include branches that are as deep as the deepest match
    # This means if you don't choose a certain branch at some point
    # the UI won't keep nagging you to choose a value for that branch
    match_tree = prune_short_branches(match_tree)

    return match_tree

@app.get("/api/paths")
async def api_paths(request: Request):
    request_dict = request_to_dict(request)
    match_tree = match_against_cache(request_dict, tree)
    match_tree = prune_short_branches(match_tree)
    paths = get_paths_to_leaves(match_tree)
    

    # deduplicate leaves based on the key
    by_path = defaultdict(lambda : {"paths" : set(), "values" : set()})
    for p in paths:
        if p[-1] == "_END_": continue
        key, values = p[-1].split("=")
        values = values.split(",")
        path = tuple(p[:-1])

        by_path[key]["values"].update(values)
        by_path[key]["paths"].add(tuple(path))

    return [{
            "paths": list(v["paths"]),
            "key": key,
            "values": sorted(v["values"], reverse=True),
    } for key, v in by_path.items()]

@app.get("/api/stac")
async def get_STAC(request: Request):
    request_dict = request_to_dict(request)
    paths = await api_paths(request)


    def make_link(key_name, paths, values):
        """Take a MARS Key and information about which paths matched up to this point and use it to make a STAC Link"""
        path = paths[0]
        href_template = f"/stac?{'&'.join(path)}{'&' if path else ''}{key_name}={{}}"
        optional = [False]
        optional_str = "Yes" if all(optional) and len(optional) > 0 else ("Sometimes" if any(optional) else "No")
        values_from_mars_language = mars_language.get(key_name, {}).get("values", [])
        
        # values = [v[0] if isinstance(v, list) else v for v in values_from_mars_language]
        
        if all(isinstance(v, list) for v in values_from_mars_language):
            value_descriptions_dict = {k : v[-1]
                                       for v in values_from_mars_language
                                       if len(v) > 1
                                       for k in v[:-1]}
            value_descriptions = [value_descriptions_dict.get(v, "") for v in values]
            if not any(value_descriptions): value_descriptions = None

        return {
                "title": key_name,
                "generalized_datacube:href_template": href_template,
                "rel": "child",
                "type": "application/json",
                "generalized_datacube:dimension" : {
                    "type" : mars_language.get(key_name, {}).get("type", ""),
                    "description": mars_language.get(key_name, {}).get("description", ""),
                    "values" : values,
                    "value_descriptions" : value_descriptions,
                    "optional" : any(optional),
                    "multiple": True,
                    "paths" : paths,
                }

            }


    def value_descriptions(key, values):
        return {
            v[0] : v[-1] for v in mars_language.get(key, {}).get("values", [])
            if len(v) > 1 and v[0] in list(values)
        }

    descriptions = {
        key : {
            "key" : key,
            "values" : values,
            "description" : mars_language.get(key, {}).get("description", ""),
            "value_descriptions" : value_descriptions(key,values),
        }
        for key, values in request_dict.items()
    }

    # Format the response as a STAC collection
    stac_collection = {
        "type": "Collection",
        "stac_version": "1.0.0",
        "id": "partial-matches",
        "description": "STAC collection representing potential children of this request",
        "links": [
            make_link(p["key"], p["paths"], p["values"])
            for p in paths
        ],
        "debug": {
            "request": request_dict,
            "descriptions": descriptions,
            "paths" : paths,
        }
    }

    return stac_collection


from frozendict import frozendict
from qubed import Qube


def test_iter_leaves_simple():
    def make_hashable(l):
        for d in l:
            yield frozendict(d)
    q = Qube.from_dict({
        "a=1/2" : {"b=1/2" : {}}
    })
    entries = [
        {"a" : '1', "b" : '1'},
        {"a" : '1', "b" : '2'},
        {"a" : '2', "b" : '1'},
        {"a" : '2', "b" : '2'},
    ]

    assert set(make_hashable(q.leaves())) == set(make_hashable(entries))

# def test_iter_leaves():
#     d = {
#         "class=od" : {
#             "expver=0001": {"param=1":{}, "param=2":{}},
#             "expver=0002": {"param=1":{}, "param=2":{}},
#         },
#         "class=rd" : {
#             "expver=0001": {"param=1":{}, "param=2":{}, "param=3":{}},
#             "expver=0002": {"param=1":{}, "param=2":{}},
#         },
#     }
#     q = Qube.from_dict(d)
#     r = Qube.from_dict(d)

#     assert q == r


import dataclasses
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from typing import Any, FrozenSet, Iterable, Literal, TypeVar


@dataclass(frozen=True)
class Values(ABC):
    @abstractmethod
    def summary(self) -> str:
        pass
    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __contains__(self, value: Any) -> bool:
        pass

    @abstractmethod
    def __iter__(self) -> Iterable[Any]:
        pass

    @abstractmethod
    def from_strings(self, values: Iterable[str]) -> list['Values']:
        pass
    
    @abstractmethod
    def min(self):
        pass

T = TypeVar("T")
EnumValuesType = FrozenSet[T]
@dataclass(frozen=True, order=True)

class QEnum(Values):
    """
    The simplest kind of key value is just a list of strings.
    summary -> string1/string2/string....
    """
    values: EnumValuesType

    def __init__(self, obj):
       object.__setattr__(self, 'values', frozenset(obj))

    def __post_init__(self):
        assert isinstance(self.values, tuple)

    def __iter__(self):
        return iter(self.values)

    def __len__(self) -> int:
        return len(self.values)
    
    def summary(self) -> str:
        return '/'.join(map(str, sorted(self.values)))
    def __contains__(self, value: Any) -> bool:
        return value in self.values
    def from_strings(self, values: Iterable[str]) -> list['Values']:
        return [type(self)(tuple(values))]
    def min(self):
        return min(self.values)

@dataclass(frozen=True)
class Range(Values, ABC):
    dtype: str = dataclasses.field(kw_only=True)

@dataclass(frozen=True)
class DateRange(Range):
    start: date
    end: date
    step: timedelta
    dtype: Literal["date"] = dataclasses.field(kw_only=True, default="date")

    def __len__(self) -> int:
        return (self.end - self.start) // self.step

    def __iter__(self) -> Iterable[date]:
        current = self.start
        while current <= self.end if self.step.days > 0 else current >= self.end:
            yield current
            current += self.step

    @classmethod
    def from_strings(self, values: Iterable[str]) -> list['DateRange']:
        dates = sorted([datetime.strptime(v, "%Y%m%d") for v in values])
        if len(dates) < 2:
            return [DateRange(
                start=dates[0],
                end=dates[0],
                step=timedelta(days=0)
            )]
        
        ranges = []
        current_range, dates = [dates[0],], dates[1:]
        while len(dates) > 1:
            if dates[0] - current_range[-1] == timedelta(days=1):
                current_range.append(dates.pop(0))
            
            elif len(current_range) == 1:
                ranges.append(DateRange(
                start=current_range[0],
                end=current_range[0],
                step=timedelta(days=0)
                ))
                current_range = [dates.pop(0),]

            else:
                ranges.append(DateRange(
                start=current_range[0],
                end=current_range[-1],
                step=timedelta(days=1)
                ))
                current_range = [dates.pop(0),]
        return ranges
    
    def __contains__(self, value: Any) -> bool:
        v = datetime.strptime(value, "%Y%m%d").date()
        return self.start <= v <= self.end and (v - self.start) % self.step == 0
    
    def summary(self) -> str:
        def fmt(d): return d.strftime("%Y%m%d")
        if self.step == timedelta(days=0):
            return f"{fmt(self.start)}"
        if self.step == timedelta(days=1):
            return f"{fmt(self.start)}/to/{fmt(self.end)}"
        
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step // timedelta(days=1)}"

@dataclass(frozen=True)
class TimeRange(Range):
    start: int
    end: int
    step: int
    dtype: Literal["time"] = dataclasses.field(kw_only=True, default="time")

    @classmethod
    def from_strings(self, values: Iterable[str]) -> list['TimeRange']:
        times = sorted([int(v) for v in values])
        if len(times) < 2:
            return [TimeRange(
                start=times[0],
                end=times[0],
                step=100
            )]
        
        ranges = []
        current_range, times = [times[0],], times[1:]
        while len(times) > 1:
            if times[0] - current_range[-1] == 1:
                current_range.append(times.pop(0))
            
            elif len(current_range) == 1:
                ranges.append(TimeRange(
                start=current_range[0],
                end=current_range[0],
                step=0
                ))
                current_range = [times.pop(0),]

            else:
                ranges.append(TimeRange(
                start=current_range[0],
                end=current_range[-1],
                step=1
                ))
                current_range = [times.pop(0),]
        return ranges

    def __len__(self) -> int:
        return (self.end - self.start) // self.step
    
    def summary(self) -> str:
        def fmt(d): return f"{d:04d}"
        if self.step == 0:
            return f"{fmt(self.start)}"
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step}"
    
    def __contains__(self, value: Any) -> bool:
        v = int(value)
        return self.start <= v <= self.end and (v - self.start) % self.step == 0

@dataclass(frozen=True)
class IntRange(Range):
    start: int
    end: int
    step: int
    dtype: Literal["int"] = dataclasses.field(kw_only=True, default="int")

    def __len__(self) -> int:
        return (self.end - self.start) // self.step
    
    def summary(self) -> str:
        def fmt(d): return d.strftime("%Y%m%d")
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step}"
    
    def __contains__(self, value: Any) -> bool:
        v = int(value)
        return self.start <= v <= self.end and (v - self.start) % self.step == 0
    
    @classmethod
    def from_strings(self, values: Iterable[str]) -> list['IntRange']:
        ints = sorted([int(v) for v in values])
        if len(ints) < 2:
            return [IntRange(
                start=ints[0],
                end=ints[0],
                step=0
            )]
        
        ranges = []
        current_range, ints = [ints[0],], ints[1:]
        while len(ints) > 1:
            if ints[0] - current_range[-1] == 1:
                current_range.append(ints.pop(0))
            
            elif len(current_range) == 1:
                ranges.append(IntRange(
                start=current_range[0],
                end=current_range[0],
                step=0
                ))
                current_range = [ints.pop(0),]

            else:
                ranges.append(IntRange(
                start=current_range[0],
                end=current_range[-1],
                step=1
                ))
                current_range = [ints.pop(0),]
        return ranges
    
def values_from_json(obj) -> Values:
    if isinstance(obj, list): 
        return QEnum(tuple(obj))

    match obj["dtype"]:
        case "date": return DateRange(**obj)
        case "time": return TimeRange(**obj)
        case "int": return IntRange(**obj)
        case _: raise ValueError(f"Unknown dtype {obj['dtype']}")



import dataclasses
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, Iterable, Literal, Sequence

from frozendict import frozendict

from . import set_operations
from .node_types import NodeData, RootNodeData
from .tree_formatters import HTML, node_tree_to_html, node_tree_to_string
from .value_types import QEnum, Values, values_from_json


@dataclass(frozen=True, eq=True, order=True)
class Qube:
    data: NodeData
    children: tuple['Qube', ...]

    @property
    def key(self) -> str:
        return self.data.key
    
    @property
    def values(self) -> Values:
        return self.data.values
    
    @property
    def metadata(self) -> frozendict[str, Any]:
        return self.data.metadata

    
    def summary(self) -> str:
        return self.data.summary()
    
    @classmethod
    def make(cls, key : str, values : Values, children, **kwargs) -> 'Qube':
        return cls(
            data = NodeData(key, values,  metadata = kwargs.get("metadata", frozendict())
            ),
            children = tuple(sorted(children, 
                                    key = lambda n : ((n.key, n.values.min()))
                                    )),
        )
    
    @classmethod
    def root_node(cls, children: Iterable["Qube"]) -> 'Qube':
        return cls.make("root", QEnum(("root",)), children)
    
    @classmethod
    def from_datacube(cls, datacube: dict[str, str | Sequence[str]]) -> 'Qube':
        key_vals = list(datacube.items())[::-1]

        children: list["Qube"] = []
        for key, values in key_vals:
            if not isinstance(values, list):
                values = [values]
            children = [cls.make(key, QEnum(values), children)]
        
        return cls.root_node(children)


    @classmethod
    def from_json(cls, json: dict) -> 'Qube':
        def from_json(json: dict) -> Qube:
            return Qube.make(
                key=json["key"],
                values=values_from_json(json["values"]),
                metadata=json["metadata"] if "metadata" in json else {},
                children=(from_json(c) for c in json["children"]),
            )
        return from_json(json)
    
    @classmethod
    def from_dict(cls, d: dict) -> 'Qube':
        def from_dict(d: dict) -> list[Qube]:
            return [
                Qube.make(
                    key=k.split("=")[0],
                    values=QEnum((k.split("=")[1].split("/"))),
                    children=from_dict(children)
                ) for k, children in d.items()]
        
        return Qube.root_node(from_dict(d))
    
    @classmethod
    def empty(cls) -> 'Qube':
        return Qube.root_node([])

    
    def __str__(self, depth = None, name = None) -> str:
        node = dataclasses.replace(self, data = RootNodeData(key = name, values=self.values, metadata=self.metadata)) if name is not None else self
        return "".join(node_tree_to_string(node=node, depth = depth))
    
    def print(self, depth = None, name: str | None = None): 
        print(self.__str__(depth = depth, name = name))
    
    def html(self, depth = 2, collapse = True, name: str | None = None) -> HTML:
        node = dataclasses.replace(self, data = RootNodeData(key = name, values=self.values, metadata=self.metadata)) if name is not None else self
        return HTML(node_tree_to_html(node=node, depth = depth, collapse = collapse))
    
    def _repr_html_(self) -> str:
        return node_tree_to_html(self, depth = 2, collapse = True)
    
    def __or__(self, other: "Qube") -> "Qube":
        return set_operations.operation(self, other, set_operations.SetOperation.UNION, type(self))
    
    def __and__(self, other: "Qube") -> "Qube":
        return set_operations.operation(self, other, set_operations.SetOperation.INTERSECTION, type(self))
    
    def __sub__(self, other: "Qube") -> "Qube":
        return set_operations.operation(self, other, set_operations.SetOperation.DIFFERENCE, type(self))
    
    def __xor__(self, other: "Qube") -> "Qube":
        return set_operations.operation(self, other, set_operations.SetOperation.SYMMETRIC_DIFFERENCE, type(self))
    
    def leaves(self) -> Iterable[dict[str, str]]:
        for value in self.values:
            if not self.children: 
                yield {self.key : value}
            for child in self.children:
                for leaf in child.leaves():
                    if self.key != "root":
                        yield {self.key : value, **leaf}
                    else:
                        yield leaf

    def datacubes(self) -> "Qube":
        def to_list_of_cubes(node: Qube) -> Iterable[Qube]:
            if not node.children:
                yield node
            # print(node.key)
            for c in node.children:
                # print(c)
                for sub_cube in to_list_of_cubes(c):
                    yield dataclasses.replace(node, children=[sub_cube])
                

        return Qube.root_node((q for c in self.children for q in to_list_of_cubes(c)))
    
    def __getitem__(self, args) -> 'Qube':
        key, value = args
        for c in self.children:
            if c.key == key and value in c.values:
                data = dataclasses.replace(c.data, values = QEnum((value,)))
                return dataclasses.replace(c, data = data)
        raise KeyError(f"Key {key} not found in children of {self.key}")

    @cached_property
    def n_leaves(self) -> int:
        # This line makes the equation q.n_leaves + r.n_leaves == (q | r).n_leaves true is q and r have no overlap
        if self.key == "root" and not self.children: return 0
        return len(self.values) * (sum(c.n_leaves for c in self.children) if self.children else 1)

    @cached_property
    def n_nodes(self) -> int:
        if self.key == "root" and not self.children: return 0
        return 1 + sum(c.n_nodes for c in self.children)

    def transform(self, func: 'Callable[[Qube], Qube | list[Qube]]') -> 'Qube':
        """
        Call a function on every node of the Qube, return one or more nodes.
        If multiple nodes are returned they each get a copy of the (transformed) children of the original node.
        Any changes to the children of a node will be ignored.
        """
        def transform(node: Qube) -> list[Qube]:
            children = [cc for c in node.children for cc in transform(c)]
            new_nodes = func(node)
            if isinstance(new_nodes, Qube):
                new_nodes = [new_nodes]

            return [dataclasses.replace(new_node, children = children)
                    for new_node in new_nodes]
        
        children = tuple(cc for c in self.children for cc in transform(c))
        return dataclasses.replace(self, children = children)

    
    def select(self, selection : dict[str, str | list[str]], mode: Literal["strict", "relaxed"] = "relaxed") -> 'Qube':
        # make all values lists
        selection = {k : v if isinstance(v, list) else [v] for k,v in selection.items()}

        def not_none(xs): return tuple(x for x in xs if x is not None)

        def select(node: Qube) -> Qube | None: 
            # Check if the key is specified in the selection
            if node.key not in selection: 
                if mode == "strict":
                    return None
                return dataclasses.replace(node, children = not_none(select(c) for c in node.children))
            
            # If the key is specified, check if any of the values match
            values = QEnum((c for c in selection[node.key] if c in node.values))

            if not values: 
                return None 
            
            data = dataclasses.replace(node.data, values = values)
            return dataclasses.replace(node, data=data, children = not_none(select(c) for c in node.children))
            
        return dataclasses.replace(self, children = not_none(select(c) for c in self.children))
    
    def span(self, key: str) -> list[str]:
        """
        Search the whole tree for any value that a given key takes anywhere.
        """
        this = set(self.values) if self.key == key else set() 
        return sorted(this | set(v for c in self.children for v in c.span(key)))
    
    def axes(self) -> dict[str, set[str]]:
        """
        Return a dictionary of all the spans of the keys in the qube.
        """
        axes = defaultdict(set)
        for c in self.children:
            for k, v in c.axes().items():
                axes[k].update(v)
        if self.key != "root":
            axes[self.key].update(self.values)
        return dict(axes)

    @staticmethod
    def _insert(position: "Qube", identifier : list[tuple[str, list[str]]]):
        """
        This algorithm goes as follows:
        We're at a particular node in the Qube, and we have a list of key-values pairs that we want to insert.
        We take the first key values pair
        key, values = identifier.pop(0)

        The general idea is to insert key, values into the current node and use recursion to handle the rest of the identifier.
        
        We have two sources of values with possible overlap. The values to insert and the values attached to the children of this node.
        For each value coming from either source we put it in one of three categories:
            1) Values that exist only in the already existing child. (Coming exclusively from position.children)
            2) Values that exist in both a child and the new values.
            3) Values that exist only in the new values.
            

        Thus we add the values to insert to a set, and loop over the children.
        For each child we partition its values into the three categories.

        For 1) we create a new child node with the key, reduced set of values and the same children.
        For 2)
            Create a new child node with the key, and the values in group 2
            Recurse to compute the children

        Once we have finished looping over children we know all the values left over came exclusively from the new values.
        So we:
            Create a new node with these values.
            Recurse to compute the children

        Finally we return the node with all these new children.
        """
        pass
        # if not identifier:
        #     return position

        # key, values = identifier.pop(0)
        # # print(f"Inserting {key}={values} into {position.summary()}")

        # # Only the children with the matching key are relevant.
        # source_children = {c : [] for c in position.children if c.key == key}
        # new_children = []

        # values = set(values)
        # for c in source_children:
        #     values_set = set(c.values)
        #     group_1 = values_set - values
        #     group_2 = values_set & values
        #     values = values - values_set # At the end of this loop values will contain only the new values

        #     if group_1:
        #         group_1_node = Qube.make(c.key, QEnum((group_1)), c.children)
        #         new_children.append(group_1_node) # Add the unaffected part of this child
            
        #     if group_2:
        #         new_node = Qube.make(key, QEnum((affected)), [])
        #         new_node = Qube._insert(new_node, identifier)
        #         new_children.append(new_node) # Add the affected part of this child


        #     unaffected = [x for x in c.values if x not in affected]


        #     if affected: # This check is not technically necessary, but it makes the code more readable


        # # If there are any values not in any of the existing children, add them as a new child
        # if entirely_new_values:
        #     new_node = Qube.make(key, QEnum((entirely_new_values)), [])
        #     new_children.append(Qube._insert(new_node, identifier))

        return Qube.make(position.key, position.values, new_children)

    def insert(self, identifier : dict[str, list[str]]) -> 'Qube':
        insertion = [(k, v) for k, v in identifier.items()]
        return Qube._insert(self, insertion)

    def info(self):
        cubes = self.to_list_of_cubes()
        print(f"Number of distinct paths: {len(cubes)}")

    @cached_property
    def structural_hash(self) -> int:
        """
        This hash takes into account the key, values and children's key values recursively.
        Because nodes are immutable, we only need to compute this once.
        """
        def hash_node(node: Qube) -> int:
            return hash((node.key, node.values, tuple(c.structural_hash for c in node.children)))

        return hash_node(self)

    def compress(self) -> "Qube":
        # First compress the children (this recursively compresses all the way to the leaves)
        new_children = [child.compress() for child in self.children]

        # Now compress the set of children at this level
        new_children = set_operations.compress_children(new_children)

        # Return the now compressed node
        return Qube(
            data = self.data,
            children = tuple(sorted(new_children))
        )


from dataclasses import dataclass, field
from typing import Hashable

from frozendict import frozendict

from .value_types import Values


@dataclass(frozen=True, eq=True, order=True)
class NodeData:
    key: str
    values: Values
    metadata: dict[str, tuple[Hashable, ...]] = field(default_factory=frozendict, compare=False)

    def summary(self) -> str:
        return f"{self.key}={self.values.summary()}" if self.key != "root" else "root"
    
@dataclass(frozen=True, eq=True, order=True)
class RootNodeData(NodeData):
    "Helper class to print a custom root name"
    def summary(self) -> str:
        return self.key


import argparse
import sys

from rich.console import Console

from qubed import Qube
from qubed.convert import parse_fdb_list

console = Console(stderr=True)


def main():
    parser = argparse.ArgumentParser(description="Generate a compressed tree from various inputs.")
    
    subparsers = parser.add_subparsers(title="subcommands", required=True)
    parser_convert = subparsers.add_parser('convert', help='Convert trees from one format to another.')
    parser_another = subparsers.add_parser('another_subcommand', help='Does something else')

    parser_convert.add_argument(
        "--input",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="Specify the input file (default: standard input)."
    )
    parser_convert.add_argument(
        "--output",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Specify the output file (default: standard output)."
    )

    parser_convert.add_argument(
        "--input_format",
        choices=["fdb", "mars"],
        default="fdb",
        help="""Specify the input format:
            fdb: the output of fdb list --porcelain
            mars: the output of mars list
        """
    )
    
    parser_convert.add_argument(
        "--output_format",
        choices=["text", "html"],
        default="text",
        help="Specify the output format (text or html)."
    )
    parser_convert.set_defaults(func=convert)
    
    args = parser.parse_args()
    args.func(args)

def convert(args):
    q = Qube.empty()
    for datacube in parse_fdb_list(args.input):
        new_branch = Qube.from_datacube(datacube)
        q = (q | Qube.from_datacube(datacube))

    # output = match args.output_format:
    #     case "text":
    #         str(q)
    #     case "html":
    #         q.html()
    output = "fw"

    with open(args.output, "w") as f:
        f.write(output)

    console.print([1, 2, 3])
    console.print("[blue underline]Looks like a link")
    console.print(locals())
    console.print("FOO", style="white on blue")

if __name__ == "__main__":
    main()



import re
import subprocess
from pathlib import Path

CARGO_TOML_PATH = Path("Cargo.toml")

# Get the latest Git tag and strip the leading 'v' if present
def get_git_version():
    try:
        version = subprocess.check_output(["git", "describe", "--tags", "--always"], text=True).strip()
        version = re.sub(r"^v", "", version)  # Remove leading 'v'
        return version
    except subprocess.CalledProcessError:
        raise RuntimeError("Failed to get Git tag. Make sure you have at least one tag in the repository.")

# Update version in Cargo.toml
def update_cargo_version(new_version):
    cargo_toml = CARGO_TOML_PATH.read_text()

    # Replace version in [package] section
    updated_toml = re.sub(r'^version = "[^"]+"', f'version = "{new_version}"', cargo_toml, flags=re.MULTILINE)

    CARGO_TOML_PATH.write_text(updated_toml)

if __name__ == "__main__":
    version = get_git_version()
    print(f"Parsed version: {version}")

    update_cargo_version(version)
    print(f"Updated Cargo.toml with version: {version}")


def parse_key_value_pairs(text: str):
    result = {}
    text = text.replace("}{", ",")  # Replace segment separators
    text = text.replace("{", "").replace("}","").strip()  # Remove leading/trailing braces

    for segment in text.split(","):
        if "=" not in segment: print(segment)
        key, values = segment.split("=", 1)  # Ensure split only happens at first "="
        values = values.split("/")
        result[key] = values

    return result

def parse_fdb_list(f):
    for line in f.readlines():
        # Handle fdb list normal
        if line.startswith("{"):
            yield parse_key_value_pairs(line)

        # handle fdb list --compact
        if line.startswith("retrieve,") and not line.startswith("retrieve,\n"):
            line = line[9:]
            yield parse_key_value_pairs(line)


import os

from flask import (
    Flask,
    render_template,
    request,
)
from flask_cors import CORS
from werkzeug.middleware.proxy_fix import ProxyFix

app = Flask(__name__)
CORS(app, resources={r"/api/*": {"origins": "*"}})

# This is required because when running in k8s the flask server sits behind a TLS proxy
# So flask speaks http while the client speaks https
# Client <-- https ---> Proxy <---- http ---> Flask server
# For the Oauth flow, flask needs to provide a callback url and it needs to use the right scheme=https
# This line tells flask to look at HTTP headers set by the TLS proxy to figure out what the original 
# Traffic looked like.
# See https://flask.palletsprojects.com/en/3.0.x/deploying/proxy_fix/
app.wsgi_app = ProxyFix(
    app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_prefix=1
)

config = {}

@app.route("/")
def index():
    return render_template("index.html", request = request, config = config, api_url = os.environ.get("API_URL", "/api/stac"))




import random
from dataclasses import dataclass
from typing import Iterable, Protocol, Sequence, runtime_checkable


@runtime_checkable
class TreeLike(Protocol):
    @property
    def children(self) -> Sequence["TreeLike"]: ...  # Supports indexing like node.children[i]
    
    def summary(self) -> str: ...

@dataclass(frozen=True)
class HTML():
    html: str
    def _repr_html_(self):
        return self.html

def summarize_node(node: TreeLike, collapse = False, **kwargs) -> tuple[str, TreeLike]:
    """
    Extracts a summarized representation of the node while collapsing single-child paths.
    Returns the summary string and the last node in the chain that has multiple children.
    """
    summaries = []
    
    while True:
        summary = node.summary(**kwargs)
        if len(summary) > 50:
            summary = summary[:50] + "..."
        summaries.append(summary)
        if not collapse: 
            break

        # Move down if there's exactly one child, otherwise stop
        if len(node.children) != 1:
            break
        node = node.children[0]

    return ", ".join(summaries), node

def node_tree_to_string(node : TreeLike, prefix : str = "", depth = None) -> Iterable[str]:
    summary, node = summarize_node(node)
    
    if depth is not None and depth <= 0:
        yield summary + " - ...\n"
        return
    # Special case for nodes with only a single child, this makes the printed representation more compact
    elif len(node.children) == 1:
        yield summary + ", "
        yield from node_tree_to_string(node.children[0], prefix, depth = depth)
        return
    else:
        yield summary + "\n"
    
    for index, child in enumerate(node.children):
        connector = "└── " if index == len(node.children) - 1 else "├── "
        yield prefix + connector
        extension = "    " if index == len(node.children) - 1 else "│   "
        yield from node_tree_to_string(child, prefix + extension, depth = depth - 1 if depth is not None else None)

def _node_tree_to_html(node : TreeLike, prefix : str = "", depth = 1, connector = "", **kwargs) -> Iterable[str]:
    summary, node = summarize_node(node, **kwargs)
    
    if len(node.children) == 0:
        yield f'<span class="leaf">{connector}{summary}</span>'
        return
    else:
        open = "open" if depth > 0 else ""
        yield f"<details {open}><summary>{connector}{summary}</summary>"

    for index, child in enumerate(node.children):
        connector = "└── " if index == len(node.children) - 1 else "├── "
        extension = "    " if index == len(node.children) - 1 else "│   "
        yield from _node_tree_to_html(child, prefix + extension, depth = depth - 1, connector = prefix+connector, **kwargs)
    yield "</details>"

def node_tree_to_html(node : TreeLike, depth = 1, **kwargs) -> str:
        css_id = f"qubed-tree-{random.randint(0, 1000000)}"
        css = f"""
        <style>
        pre#{css_id} """ \
        """{
            font-family: monospace;
            white-space: pre;
            font-family: SFMono-Regular,Menlo,Monaco,Consolas,Liberation Mono,Courier New,Courier,monospace;
            font-size: 12px;
            line-height: 1.4;
            
            details {
                margin-left: 0;
            }

            summary {
                list-style: none;
                cursor: pointer;
                text-overflow: ellipsis;
                overflow: hidden;
                text-wrap: nowrap;
                display: block;
            }

            summary:hover,span.leaf:hover {
                background-color: #f0f0f0;
            }

            details > summary::after {
                content: ' ▲';
            }

            details:not([open]) > summary::after {
                content: " ▼";
            }

            .leaf {
                text-overflow: ellipsis;
                overflow: hidden;
                text-wrap: nowrap;
                display: block;
            }

            summary::-webkit-details-marker {
              display: none; 
              content: "";
            }

        }
        </style>
        """
        nodes = "".join(_node_tree_to_html(node=node, depth=depth, **kwargs))
        return f"{css}<pre class='qubed-tree' id='{css_id}'>{nodes}</pre>"


from tree_traverser import backend, CompressedTree
import datetime
import psutil
from tqdm import tqdm
from pathlib import Path
import json
from more_itertools import chunked
process = psutil.Process()

def massage_request(r):
    return {k : v if isinstance(v, list) else [v]
            for k, v in r.items()}


if __name__ == "__main__":

    config = """
---
type: remote
host: databridge-prod-catalogue1-ope.ewctest.link
port: 10000
engine: remote
store: remote
    """

    request = {
            "class": "d1",
            "dataset": "climate-dt",
            # "date": "19920420",
        }
    
    data_path = Path("data/compressed_tree_climate_dt.json")
    if not data_path.exists():
        compressed_tree = CompressedTree({})
    else:
        compressed_tree = CompressedTree.load(data_path)

    fdb = backend.PyFDB(fdb_config = config)

    visited_path = Path("data/visited_dates.json")
    if not visited_path.exists():
        visited_dates = set()
    else:
        with open(visited_path, "r") as f:
            visited_dates = set(json.load(f))

    today = datetime.datetime.today()
    start = datetime.datetime.strptime("19920420", "%Y%m%d")
    date_list = [start + datetime.timedelta(days=x) for x in range((today - start).days)]
    date_list = [d.strftime("%Y%m%d") for d in date_list if d not in visited_dates] 
    for dates in chunked(tqdm(date_list), 5):
        print(dates[0])
        print(f"Memory usage: {(process.memory_info().rss)/1e6:.1f} MB")

        r = request | dict(date = dates)
        tree = fdb.traverse_fdb(massage_request(r))

        compressed_tree.insert_tree(tree)
        compressed_tree.save(data_path)
        
        for date in dates:
            visited_dates.add(date)

        with open(visited_path, "w") as f:
            json.dump(list(visited_dates), f)
        
        # print(compressed_tree.reconstruct_compressed_ecmwf_style())



from collections import defaultdict
from dataclasses import replace
from enum import Enum

# Prevent circular imports while allowing the type checker to know what Qube is
from typing import TYPE_CHECKING, Iterable

from frozendict import frozendict

from .node_types import NodeData
from .value_types import QEnum, Values

if TYPE_CHECKING:
    from .qube import Qube


class SetOperation(Enum):
    UNION = (1, 1, 1)
    INTERSECTION = (0, 1, 0)
    DIFFERENCE = (1, 0, 0)
    SYMMETRIC_DIFFERENCE = (1, 0, 1)

def fused_set_operations(A: "Values", B: "Values") -> tuple[list[Values], list[Values], list[Values]]:
    if isinstance(A, QEnum) and isinstance(B, QEnum):
        set_A, set_B = set(A), set(B)
        intersection = set_A & set_B
        just_A = set_A - intersection
        just_B = set_B - intersection
        return [QEnum(just_A),], [QEnum(intersection),], [QEnum(just_B),]
                
    
    raise NotImplementedError("Fused set operations on values types other than QEnum are not yet implemented")

def node_intersection(A: "Values", B: "Values") -> tuple[Values, Values, Values]:
    if isinstance(A, QEnum) and isinstance(B, QEnum):
        set_A, set_B = set(A), set(B)
        intersection = set_A & set_B
        just_A = set_A - intersection
        just_B = set_B - intersection
        return QEnum(just_A), QEnum(intersection), QEnum(just_B)
                
    
    raise NotImplementedError("Fused set operations on values types other than QEnum are not yet implemented")

def operation(A: "Qube", B : "Qube", operation_type: SetOperation, node_type) -> "Qube":
    assert A.key == B.key, "The two Qube root nodes must have the same key to perform set operations," \
                           f"would usually be two root nodes. They have {A.key} and {B.key} respectively"
    
    assert A.values == B.values, f"The two Qube root nodes must have the same values to perform set operations {A.values = }, {B.values = }"

    # Group the children of the two nodes by key
    nodes_by_key = defaultdict(lambda : ([], []))
    for node in A.children:
        nodes_by_key[node.key][0].append(node)
    for node in B.children:
        nodes_by_key[node.key][1].append(node)

    new_children = []

    # For every node group, perform the set operation
    for key, (A_nodes, B_nodes) in nodes_by_key.items():
        new_children.extend(_operation(key, A_nodes, B_nodes, operation_type, node_type))

    # Whenever we modify children we should recompress them
    # But since `operation` is already recursive, we only need to compress this level not all levels
    # Hence we use the non-recursive _compress method
    new_children = compress_children(new_children)

    # The values and key are the same so we just replace the children
    return replace(A, children=new_children)
    

# The root node is special so we need a helper method that we can recurse on
def _operation(key: str, A: list["Qube"], B : list["Qube"], operation_type: SetOperation, node_type) -> Iterable["Qube"]:
    # We need to deal with the case where only one of the trees has this key.
    # To do so we can insert a dummy node with no children and no values into both A and B
    keep_just_A, keep_intersection, keep_just_B = operation_type.value

    # Iterate over all pairs (node_A, node_B)
    values = {}
    for node in A + B:
        values[node] = node.values

    for node_a in A:
        for node_b in B:

            # Compute A - B, A & B, B - A
            # Update the values for the two source nodes to remove the intersection
            just_a, intersection, just_b = node_intersection(
                values[node_a], 
                values[node_b], 
            )

            # Remove the intersection from the source nodes
            values[node_a] = just_a
            values[node_b] = just_b

            if keep_intersection:
                if intersection:
                    new_node_a = replace(node_a, data = replace(node_a.data, values = intersection))
                    new_node_b = replace(node_b, data= replace(node_b.data, values = intersection))
                    yield operation(new_node_a, new_node_b, operation_type, node_type)


    # Now we've removed all the intersections we can yield the just_A and just_B parts if needed
    if keep_just_A:
        for node in A:
            if values[node]:
                yield node_type.make(key, values[node], node.children)
    if keep_just_B:
        for node in B:
            if values[node]:
                yield node_type.make(key, values[node], node.children)

def compress_children(children: Iterable["Qube"]) -> tuple["Qube"]:
    """
    Helper method tht only compresses a set of nodes, and doesn't do it recursively.
    Used in Qubed.compress but also to maintain compression in the set operations above.
    """
    # Now take the set of new children and see if any have identical key, metadata and children
    # the values may different and will be collapsed into a single node
    identical_children = defaultdict(set)
    for child in children:
        # only care about the key and children of each node, ignore values
        key = hash((child.key, tuple((cc.structural_hash for cc in child.children))))
        identical_children[key].add(child)
    
    # Now go through and create new compressed nodes for any groups that need collapsing
    new_children = []
    for child_set in identical_children.values():
        if len(child_set) > 1:
            child_set = list(child_set)
            node_type = type(child_set[0])
            key = child_set[0].key

            # Compress the children into a single node
            assert all(isinstance(child.data.values, QEnum) for child in child_set), "All children must have QEnum values"
            
            node_data = NodeData(
                key = key,
                metadata = frozendict(), # Todo: Implement metadata compression
                values = QEnum((v for child in child_set for v in child.data.values.values)),
            )
            new_child = node_type(data = node_data, children = child_set[0].children)
        else:
            # If the group is size one just keep it
            new_child = child_set.pop()
        
        new_children.append(new_child)
    return tuple(sorted(new_children, 
                        key = lambda n : ((n.key, tuple(sorted(n.values.values))))
                        ))



from qubed import Qube


def test_leaf_conservation():
    q = Qube.from_dict({
        "class=d1": {"dataset=climate-dt" : {  
            "time=0000": {"param=130/134/137/146/147/151/165/166/167/168/169" : {}},
            "time=0001": {"param=130": {}},
                    }}})

    r  = Qube.from_datacube({
        "class": "d1",
        "dataset": "climate-dt",
        "time": "0001",
        "param": "134"
    })

    assert q.n_leaves + r.n_leaves == (q | r).n_leaves